/*
 * anim.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "anim.hpp"
#include "binary_io/binary_io.hpp"
#include "cfg.hpp"
#include "client_config.hpp"
#include "misc/rng.hpp"
#include "graphic.hpp"

const Graphic * Anim::getGraphic(MapDirection facing, int frame, int x, int y,
								 int gvt) const
{
	if (vbat_mode) {
		if (frame != 0) {
			// "Biting" frame
			return g[D_WEST][0];
		} else {
			// "Flapping" frame -- this has to be chosen randomly.

			static const int bat_anim_timescale = CfgInt("bat_anim_timescale");
			static int last_gvt = -5000;
			static int frames[64];

			if (gvt > last_gvt + (bat_anim_timescale<<7)) {
				// (every 2**7 == 128 timescales, we regenerate the random table.)
				for (int i=0; i<64; ++i) {
					frames[i] = g_rng.getInt(0,3);
				}
				last_gvt = gvt;
			}

			// The exact graphic is chosen based on gvt, facing and
			// mc. This should sufficiently randomize things. (Note we
			// don't just switch graphics each frame; that would be
			// way too fast. The procedure here is designed to make
			// the same graphic remain for a fixed time before we
			// switch to a new, randomly chosen graphic.)
			int idx = gvt / bat_anim_timescale + int(facing) + (x<<3) + y;
			idx >>= 2;
			idx = idx & 63;
			return g[frames[idx]][0];
		}
		
	} else {
		// For most anims we just do a straight lookup, w/o randomizing anything!
		return g[facing][frame];
	}
}

void Anim::serialize(ostream &str, int new_id)
{
	id = new_id;
	for (int i=0; i<4; ++i) {
		for (int j=0; j<NFRAMES; ++j) {
			if (g[i][j]) BinWriteUword(str, GetID(*g[i][j]));
			else BinWriteUword(str, 0);
		}
	}
	cc_normal.serialize(str);
	cc_invulnerable.serialize(str);
	BinWriteUbyte(str, vbat_mode?1:0);
}

Anim::Anim(istream &str, const ClientConfig &conf)
	: id(0), vbat_mode(false)
{
	for (int i=0; i<4; ++i) {
		for (int j=0; j<NFRAMES; ++j) {
			g[i][j] = conf.getGraphic(BinReadUword(str));
		}
	}
	cc_normal = ColourChange(str);
	cc_invulnerable = ColourChange(str);
	const int tmp = BinReadUbyte(str);
	vbat_mode = (tmp != 0);
}



namespace {
	struct OffsetData {
		MapDirection dir;
		int ofsx, ofsy;
	};
	OffsetData offset_data[N_OVERLAY];
}

void GetOverlayOffset(MapDirection facing, int frame, int &ofsx, int &ofsy, MapDirection &dir, bool &show)
{
	show = (frame != AF_THROW_DOWN);
	
	if (frame >= 0 && frame <= 4) {
		const OffsetData &od(offset_data[frame*4+facing]);
		ofsx = od.ofsx;
		ofsy = od.ofsy;
		dir = od.dir;
	} else {
		ofsx = ofsy = 0;
		dir = D_NORTH;
	}
}


void Overlay::serialize(ostream &str, int new_id)
{
	id = new_id;
	for (int i=0; i<4; ++i) {
		if (g[i]) BinWriteUword(str, GetID(*g[i]));
		else BinWriteUword(str, 0);
	}
}

Overlay::Overlay(istream &str, const ClientConfig &conf)
	: id(0)
{
	for (int i=0; i<4; ++i) {
		g[i] = conf.getGraphic(BinReadUword(str));
	}
}

void InitOverlayOffsets(const int *p)
{
	for (int i=0; i<N_OVERLAY; ++i) {
		int dir = *p++;
		if (dir < 0 || dir > 3) dir = 0;
		offset_data[i].dir = MapDirection(dir);
		offset_data[i].ofsx = *p++;
		offset_data[i].ofsy = *p++;
	}
}

void SerializeOverlayOffsets(ostream &str)
{
	for (int i=0; i<N_OVERLAY; ++i) {
		BinWriteInt(str, offset_data[i].dir);
		BinWriteInt(str, offset_data[i].ofsx);
		BinWriteInt(str, offset_data[i].ofsy);
	}
}

void InitOverlayOffsets(istream &str)
{
	int x[N_OVERLAY*3];
	for (int i=0; i<N_OVERLAY*3; ++i) x[i] = BinReadInt(str);
	InitOverlayOffsets(x);
}
