/*
 * display.hpp
 *
 * Display: represents a window or display device. Contains operations for drawing on the display
 * (abstraction of the drawing operations of SDL or DirectDraw or whatever).
 *
 * Uses a pimpl so that the details of the graphics library being used are not exposed to the rest
 * of the program. Most of the Display routines are actually implemented in files under platform/
 * (e.g. platform/sdl_display.cpp). Porting to a new platform should be relatively easy.
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#ifndef DISPLAY_HPP
#define DISPLAY_HPP

#include "gfx_resizer.hpp"

#include "boost/shared_ptr.hpp"
using namespace boost;

class ColourChange;
class DisplayImpl;
class Graphic;

class Display {
public:
    // ctor (NB no dtor is needed, this can be placed in the display_impl dtor if necessary.)
    Display(const shared_ptr<GfxResizer> resizer) : gfx_resizer(resizer) { construct(); }

    // Get/set GfxResizer
    void setGfxResizer(shared_ptr<GfxResizer> resizer) { gfx_resizer = resizer; onChangeSizer(); }
    shared_ptr<GfxResizer> getGfxResizer() const { return gfx_resizer; }
    
    // Get current display size
    int getWidth() const;
    int getHeight() const;

    // begin: clears the screen (to the given bg colour)
    // end: flips the buffers.
    void begin(int r=0, int g=0, int b=0);
    void end();

    // clipping: rectangle is BL inclusive and TR exclusive
    // (TODO) may be better to have a sentry object for clipping, because I always
    // forget to turn it off after using it :)
    void setClipRectangle(int bl_x, int bl_y, int tr_x, int tr_y);
    void clearClipRectangle();  // turns off clipping

    // text: draw text string with BL corner at a certain point
    void setFontSize(int size);
    void drawText(int bl_x, int bl_y, const char *str);
    void getTextSize(const char *str, int &w, int &h);
    int getTextWidth(const char *str) { int w,h; getTextSize(str,w,h); return w; }
    int getTextHeight(const char *str) { int w,h; getTextSize(str,w,h); return h; }    // actual height of a piece of text.
    int getTextHeight();                   // suggested inter-line spacing for text.

    
    // draw a solid alpha-blended rectangle
    // (colours in range 0-255)
    void drawBlendedRectangle(int bl_x, int bl_y, int width, int height, int r, int g, int b, int a);
    
    // graphics: draw a Graphic with BL corner at a certain point
    void drawGraphic(int bl_x, int bl_y, const Graphic &gfx, int width, int height);

    // this draws a graphic with some of the colours changed.
    void drawGraphicWithCC(int bl_x, int bl_y, const Graphic &, int width, int height, const ColourChange &);

    // pixel plotting: (r, g and b range from 0 to 255.)
    void beginPixels();  // locks the display so pixels can be plotted or read
    void drawPixel(int x, int y, int r, int g, int b);
    void readPixel(int x, int y, int &r, int &g, int &b);
    void endPixels();    // unlocks the display

private:
    void construct();
    void onChangeSizer();
    
private:
    shared_ptr<DisplayImpl> pimpl;
    shared_ptr<GfxResizer> gfx_resizer;
};

#endif
