/*
 * enet_stream.cpp
 *
 * (TODO) Some of this code is also being used in wce_stream.cpp -- should probably refactor
 * into a common base class.
 * 
 * Copyright (c) Stephen Thompson, 2007.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "enet_controller.hpp"
#include "enet_stream.hpp"

EnetStreambuf::int_type EnetStreambuf::overflow(EnetStreambuf::int_type meta)
{
	if (!net) return traits_type::eof();
	
	if (meta != traits_type::eof()) {
		char c = traits_type::to_char_type(meta);
		if (pptr() == epptr()) {
			// We must grow the output buffer. We will double the size each time.
			const int old_size = output_buffer.size();
			const int new_size = old_size==0 ? 1024 : old_size*2;
			output_buffer.resize(new_size);
			setp(&*(output_buffer.begin()), &*(output_buffer.end()));
			pbump(old_size);
		}
		// Now write the character to the buffer.
		*(pptr()) = c;
		pbump(1);
		
		return traits_type::not_eof(meta);
	} else {
		return meta;
	}
}

EnetStreambuf::int_type EnetStreambuf::underflow()
{
	// Check whether connection is open -- if not, return eof.
	if (!net) return traits_type::eof();
	
	// Make sure we are up to date with incoming packets:
	net->processMessages();

	// Now, make sure that the input buffer is properly set
	streamsize result = updateInputBuffer();
	if (result == 0) {
		// Uh-oh -- Ran out of data. Return EOF.
		return traits_type::eof();
	} else {
		// Data byte is available -- return it
		return traits_type::to_int_type(*gptr());
	}
}

streamsize EnetStreambuf::updateInputBuffer()
{
	// Helper function used by underflow() and showmanyc()

	streamsize result;
	while (1) {
		if (input_buffer.empty()) {
			// No packets left!
			setg(0,0,0);
			return 0;
		}
		result = egptr() - gptr();
		if (result != 0) return result;  // data available in first packet -- we are done
		// Front packet has been used up.
		// Delete it, and reset the get pointers.
		enet_packet_destroy(input_buffer.front());
		input_buffer.pop_front();
		if (!input_buffer.empty()) {
			// Reset the get pointers to point to the (new) front packet.
			// NB we assume char* and enet_uint8* are interchangeable.
			char *p = reinterpret_cast<char*>(input_buffer.front()->data);
			setg(p, p, p+input_buffer.front()->dataLength);
		}
	}
}

int EnetStreambuf::sync()
{
	// Check if connection is open 
	if (!net || !peer) return -1;  // error

	// Clear out the message queue
	net->processMessages();
	
	// If the output buffer is empty, we can exit at this point
	if (pptr() == pbase()) return 0; // success
	
	// Now, write out a packet
	ENetPacket *packet = enet_packet_create(pbase(), pptr() - pbase(),
											ENET_PACKET_FLAG_RELIABLE);
	enet_peer_send(peer, 0, packet);
	// (Note that we must not call enet_packet_destroy)

	// Clear out the message queue again. This ensures that the packet is actually sent.
	net->processMessages();

	// Reset the put pointers:
	setp(&*(output_buffer.begin()), &*(output_buffer.end()));
	
	return 0; // success
}

streamsize EnetStreambuf::showmanyc()
{
	if (!net) return 0;
	
	// First, ensure that we have copied in all pending packets
	net->processMessages();

	// Now clean up the input buffer, and find the number of bytes remaining on the
	// first available packet...
	return updateInputBuffer();
}

void EnetStreambuf::receivePacket(ENetPacket *packet)
{
	input_buffer.push_back(packet);
	if (input_buffer.size()==1) {
		// Set get-pointers appropriately
		ASSERT(gptr()==0);
		char *p = reinterpret_cast<char*>(input_buffer.front()->data);
		setg(p, p, p+input_buffer.front()->dataLength);
	}
}

EnetStreambuf::~EnetStreambuf()
{
	clearInputBuffer();
}

void EnetStreambuf::close()
{
	clearInputBuffer();
	net = 0;
	peer = 0;
}

void EnetStreambuf::clearInputBuffer()
{
	// Destroy all held packets
	for (deque<ENetPacket*>::iterator it = input_buffer.begin();
	it != input_buffer.end(); ++it) {
		enet_packet_destroy(*it);
	}
	input_buffer.clear();
}
