/*
 * game_client.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "client_config.hpp"
#include "command_codes.hpp"
#include "control.hpp"
#include "game_client.hpp"
#include "local_dungeon_view.hpp"
#include "misc/my_exceptions.hpp"
#include "colour_change.hpp"

GameClient::GameClient(iostream &s, LocalDungeonView &d, ClientConfig &config)
    : str(s), dview(d), conf(config), gvt(0), last_sent_ctrl(0), game_over(false),
      game_quit(false)
{
}


//
// "recv" functions
//

void GameClient::recvNibbles(int &x, int &y)
{
    int z = BinReadUbyte(str);
    x = (z >> 4);
    y = (z & 0xf);
}

void GameClient::recvTileInfo(int &d, int &c, bool &cc)
{
    recvNibbles(d, c);
    d -= 8;
    cc = ((c&1)==1);
    c >>= 1;
}

const Anim * GameClient::recvAnimPtr()
{
    return conf.getAnim(BinReadUbyte(str));
}

const Control * GameClient::recvControlPtr()
{
    return conf.getControl(BinReadUbyte(str));
}

pair<const Control *, bool> GameClient::recvControlPlusBool()
{
    int id = BinReadUbyte(str);
    bool b = false;
    if (id >= 128) {
        b = true;
        id -= 128;
    }
    return make_pair(conf.getControl(id), b);
}

const Graphic * GameClient::recvGraphicPtr()
{
    return conf.getGraphic(BinReadUword(str));
}

const Overlay * GameClient::recvOverlayPtr()
{
    return conf.getOverlay(BinReadUbyte(str));
}

string GameClient::recvString()
{
    return BinReadString(str);
}


int GameClient::recvCommandNo()
{
    return BinReadUbyte(str);
}

MapCoord GameClient::recvMapCoord()
{
    int x = BinReadUbyte(str);
    if (x==255) return MapCoord();
    int y = BinReadUbyte(str);
    return MapCoord(x,y);
}

void GameClient::sendGameOverMsg()
{
    // 255 is the magic number that confirms to the server that the game has ended.
    // (Note this also appears in RemoteDungeonView::readController and InGameScreen::update.)
    // Please note that GameClient should *never* send a 255 byte except to signify
    // the game over state!! (otherwise InGameScreen on the server may misinterpret it as
    // a game over signal, and then things will go pear-shaped.)
    BinWriteUbyte(str, 255);
    str.flush();
}

void GameClient::sendControl(const Control *ctrl)
{
    if (ctrl) {
        if (ctrl->getID() == 0 || ctrl->getID() == 255) {
            // Control IDs are not supposed to be zero.
            // Also, make sure we are not sending the dreaded 255 byte...
            // (we shouldn't have more than 254 controls, anyway...)
            throw UnexpectedError("bad Control ID in GameClient::sendControl");
        }
        BinWriteUbyte(str, ctrl->getID());
    } else {
        BinWriteUbyte(str, 0);  // zero means "cancel the last control".
    }
}


//
// update
//

void GameClient::update(int dt)
{
    gvt += dt;
    
    // Read in any network commands.
    receiveCommands();

    // Read the controller, and send result over the network.
    // (Only if game hasn't ended though!)
    if (!game_over) {
        const Control *ctrl = dview.readController();
        if (ctrl) {
            if (ctrl->isContinuous()) {
                // Continuous controls are only sent once.
                if (ctrl != last_sent_ctrl) {
                    sendControl(ctrl);
                    last_sent_ctrl = ctrl;
                }
            } else {
                // Discts controls are sent any number of times.
                sendControl(ctrl);
                last_sent_ctrl = 0;
            }
        } else {
            if (last_sent_ctrl && last_sent_ctrl->isContinuous()) {
                sendControl(0);  // cancel last sent ctrl
            }
            last_sent_ctrl = 0;
        }
    }
    str.flush();
}



//
// receiveCommands -- translates network commands into DungeonView calls.
//

void GameClient::receiveCommands()
{
    while (str.rdbuf()->in_avail() > 0) {

        // If the game has ended, then don't try to read anything further
        // from the stream!
        if (game_over) return;

        int cmd = recvCommandNo();
        switch (cmd) {
        case C_SET_CURRENT_ROOM:
            {
                int r = recvRoomNo();
                int w, h;
                recvRoomCoord(w, h);
                w++;
                h++;
                dview.setCurrentRoom(r, w, h);
            }
            break;
            
        case C_MAP_CURRENT_ROOM:
            {
                MapCoord mc = recvMapCoord();
                dview.mapCurrentRoom(mc);
            }
            break;
            
        case C_MAP_WALL:
            {
                MapCoord mc = recvMapCoord();
                dview.mapWall(mc);
            }
            break;
            
        case C_MAP_KNIGHT_LOCATION:
            {
                int n = BinReadUbyte(str);
                MapCoord mc = recvMapCoord();
                dview.mapKnightLocation(n, mc);
            }
            break;

        case C_MAP_ITEM_LOCATION:
            {
                MapCoord mc = recvMapCoord();
                int i = BinReadUbyte(str);
                dview.mapItemLocation(mc, i);
            }
            break;
            
        case C_WIPE_MAP:
            dview.wipeMap();
            break;
            
        case C_ADD_ENTITY:
            {
                unsigned short int id = BinReadUword(str);
                int x, y;
                recvRoomCoord(x, y);
                int height, facing;
                recvNibbles(height, facing);
                const Anim *anim = recvAnimPtr();
                const Overlay *ovr = recvOverlayPtr();
                int af, xyz;
                recvNibbles(af, xyz);
                int mt = ((xyz & 2) >> 1);
                bool ainvuln = (xyz & 1);
                int atz_diff = (af==0) ? 0 : BinReadWord(str);
                int cur_ofs = BinReadUword(str);
                int mtr = (mt == MT_NOT_MOVING) ? 0 : BinReadUword(str);
                dview.addEntity(id, x, y, MapHeight(height), MapDirection(facing), anim, ovr, af,
                                atz_diff, ainvuln, cur_ofs, MotionType(mt), mtr);
            }
            break;
            
        case C_RM_ENTITY:
            {
                unsigned short int id = BinReadUword(str);
                dview.rmEntity(id);
            }
            break;
            
        case C_REPOSITION_ENTITY:
            {
                unsigned short int id = BinReadUword(str);
                int new_x, new_y;
                recvRoomCoord(new_x, new_y);
                dview.repositionEntity(id, new_x, new_y);
            }
            break;
            
        case C_MOVE_ENTITY:
            {
                unsigned short int id = BinReadUword(str);
                int mt, mm;
                recvNibbles(mt, mm);
                int motion_duration = BinReadUword(str);
                dview.moveEntity(id, MotionType(mt), motion_duration, mm);
            }
            break;
            
        case C_SET_ANIM_DATA:
            {
                unsigned short int id = BinReadUword(str);
                const Anim *anim = recvAnimPtr();
                const Overlay *ovr = recvOverlayPtr();
                int af, xyz;
                recvNibbles(af, xyz);
                bool ainvuln = ((xyz / 2) != 0);
                bool during_motion = ((xyz & 1) != 0);
                int atz_diff = BinReadWord(str);
                dview.setAnimData(id, anim, ovr, af, atz_diff, ainvuln, during_motion);
            }
            break;
            
        case C_SET_FACING:
            {
                unsigned short int id = BinReadUword(str);
                int dummy, new_facing;
                recvNibbles(dummy, new_facing);
                dview.setFacing(id, MapDirection(new_facing));
            }
            break;
                
        case C_CLEAR_TILES:
            {
                int x, y;
                recvRoomCoord(x, y);
                dview.clearTiles(x, y);
            }
            break;

        case C_SET_TILE:
            {
                int x, y;
                recvRoomCoord(x, y);
                int depth, c;
                bool cc;
                recvTileInfo(depth, c, cc);
                const Graphic *gfx = recvGraphicPtr();
                shared_ptr<const ColourChange> newcc;
                if (cc) {
                    newcc.reset(new ColourChange(str));
                }
                dview.setTile(x, y, depth, gfx, newcc, MiniMapColour(c));
            }
            break;

        case C_SET_ITEM:
            {
                int x, y;
                recvRoomCoord(x, y);
                const Graphic *g = recvGraphicPtr();
                dview.setItem(x, y, g);
            }
            break;

        case C_PLACE_ICON:
            {
                int x, y;
                recvRoomCoord(x, y);
                const Graphic *g = recvGraphicPtr();
                int dur = BinReadUword(str);
                dview.placeIcon(x, y, g, dur);
            }
            break;

        case C_SET_BACKPACK:
            {
                int slot = BinReadUbyte(str);
                const Graphic *g = recvGraphicPtr();
                const Graphic *overdraw = recvGraphicPtr();
                int no_carried = BinReadUbyte(str);
                int no_max = BinReadUbyte(str);
                dview.setBackpack(slot, g, overdraw, no_carried, no_max);
            }
            break;

        case C_ADD_SKULL:
            dview.addSkull();
            break;

        case C_SET_HEALTH:
            {
                int h = BinReadUbyte(str);
                dview.setHealth(h);
            }
            break;  
            
        case C_SET_POTION_MAGIC:
            {
                int x = BinReadUbyte(str);
                bool pi = (x & 128);
                int pm = (x & 127);
                dview.setPotionMagic(PotionMagic(pm), pi);
            }
            break;

        case C_FLASH_MESSAGE:
            {
                int ntimes = BinReadUbyte(str);
                string msg = recvString();
                dview.flashMessage(msg, ntimes);
            }
            break;

        case C_CANCEL_CONTINUOUS_MESSAGES:
            dview.cancelContinuousMessages();
            break;

        case C_ADD_CONTINUOUS_MESSAGE:
            {
                string msg = recvString();
                dview.addContinuousMessage(msg);
            }
            break;

        case C_FLASH_SCREEN:
            {
                const int delay = BinReadUword(str);
                dview.flashScreen(delay);
            }
            break;

        case C_SET_AVAILABLE_CONTROLS:
            {
                vector<pair<const Control*, bool> > ctrls;
                int sz = BinReadUbyte(str);
                ctrls.reserve(sz);
                for (int i=0; i<sz; ++i) {
                    ctrls.push_back(recvControlPlusBool());
                }
                dview.setAvailableControls(ctrls);
            }
            break;

        case C_SET_MENU_HIGHLIGHT:
            {
                const Control *ctrl = recvControlPtr();
                dview.setMenuHighlight(ctrl);
            }
            break;

        case C_WIN_GAME:
            dview.winGame();
            game_over = true;
            sendGameOverMsg();
            break;

        case C_LOSE_GAME:
            dview.loseGame();
            game_over = true;
            sendGameOverMsg();
            break;

        case C_QUIT_GAME:
            dview.quitGame();
            game_over = true;
            game_quit = true;
            sendGameOverMsg();
            break;

        case C_PLAY_SOUND:
            {
                string name = recvString();
                int freq = BinReadLong(str);
                dview.playSound(name, freq);
            }
            break;

        case C_FLIP_ENTITY_MOTION:
            {
                unsigned short int id = BinReadUword(str);
                int initial_delay = BinReadUword(str); 
                int motion_duration = BinReadUword(str);
                dview.flipEntityMotion(id, initial_delay, motion_duration);
            }
            break;
            
        default:
            throw UnexpectedError("bad command on network");
            break;
        }
    }   
}
