/*
 * game_server.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "dungeon_generator.hpp"
#include "dungeon_view.hpp"
#include "event_manager.hpp"
#include "game_server.hpp"
#include "gore_manager.hpp"
#include "home_manager.hpp"
#include "item_respawn_task.hpp"
#include "magic_map.hpp"
#include "mediator.hpp"
#include "menu.hpp"
#include "menu_selections.hpp"
#include "monster_manager.hpp"
#include "monster_task.hpp"
#include "my_exceptions.hpp"
#include "player.hpp"
#include "server_config.hpp"
#include "stuff_bag.hpp"
#include "task_manager.hpp"
#include "view_manager.hpp"

#include "boost/scoped_ptr.hpp"
using namespace boost;

//
// pimpl
//

struct GameServerImpl {

    explicit GameServerImpl(int nplyrs) : view_manager(nplyrs) { }
    
    shared_ptr<DungeonMap> dmap;
    EventManager event_manager;
    GoreManager gore_manager;
    HomeManager home_manager;
    MonsterManager monster_manager;
    StuffManager stuff_manager;
    TaskManager task_manager;
    ViewManager view_manager;
    bool premapped;
    vector<pair<const ItemType *, vector<int> > > gears;
};


//
// readMenu helper function
//

void GameServer::readMenu(const Menu &menu, const MenuSelections &ms,
                          DungeonGenerator &dg, vector<shared_ptr<Quest> > &quests)
{
    quests.clear();

    // "Global" dungeon directives
    for (int i=0; i<menu.getNumDungeonDirectives(); ++i) {
        menu.getDungeonDirective(i)->apply(dg, ms);
    }

    // MenuItem-specific dungeon directives
    for (int i=0; i<menu.getNumItems(); ++i) {
        const MenuItem &mi( menu.getItem(i) );
        const MenuValue &mv( mi.getValue(ms.get(mi.getID())) );
        for (int j=0; j<mv.getNumDungeonDirectives(); ++j) {
            mv.getDungeonDirective(j)->apply(dg, ms);
        }
        for (int j=0; j<mv.getNumQuests(); ++j) {
            // For now, same quests are given to each player.
            quests.push_back(mv.getQuest(j));
        }
    }
}


//
// initialization routines, dtor
//

GameServer::GameServer(int nplyrs)
    : pimpl(new GameServerImpl(nplyrs))
{
    pimpl->premapped = false;
}

void GameServer::initialize(const ServerConfig &sc, const MenuSelections &ms,
                            const vector<DungeonView*> dviews)
{
    // Create the mediator
    Mediator::createInstance(pimpl->event_manager, pimpl->gore_manager, pimpl->home_manager,
                             pimpl->monster_manager, pimpl->stuff_manager,
                             pimpl->task_manager, pimpl->view_manager);

    // Generate dungeon map; get quests
    scoped_ptr<DungeonGenerator> dg;
    vector<shared_ptr<Quest> > quests;
    dg.reset(new DungeonGenerator(sc.getSegmentSet(), sc.getWall(), sc.getHorizDoor(0),
                                  sc.getHorizDoor(1), sc.getVertDoor(0), sc.getVertDoor(1)));
    quests.clear();
    readMenu(sc.getMenu(), ms, *dg, quests);  // apply DungeonDirectives and get quests
    pimpl->dmap.reset(new DungeonMap);
    Mediator::instance().setMap(pimpl->dmap);
    dg->generate(*pimpl->dmap);

    // Add homes to the home manager
    for (int i=0; i<dg->getNumHomesOverall(); ++i) {
        MapCoord pos;
        MapDirection facing_toward_home;
        dg->getHomeOverall(i, pos, facing_toward_home);
        pimpl->home_manager.addHome(pos, facing_toward_home);
    }
    
    // Add players
    ASSERT(dviews.size()==2);  // (TODO) temporary. support multiple players in future.
    if (dg->getNumHomes() < dviews.size()) {
        throw UnexpectedError("dungeon generator not set up for this many players");
    }

    MapCoord pos;
    MapDirection facing;
    
    dg->getHome(0, pos, facing);
    Player *p1 = new Player(*pimpl->dmap, pos, facing,
                            sc.getKnightAnim(0), sc.getDefaultItem(),
                            0, sc.getControlSet(), *dviews[0], quests,
                            sc.getSecuredCC(0));

    dg->getHome(1, pos, facing);
    Player *p2 = new Player(*pimpl->dmap, pos, facing,
                            sc.getKnightAnim(1), sc.getDefaultItem(),
                            0, sc.getControlSet(), *dviews[1], quests,
                            sc.getSecuredCC(1));

    switch (dg->getExitType()) {
    case E_SELF:
        p1->setExitFromPlayersHome(*p1);
        p2->setExitFromPlayersHome(*p2);
        break;
    case E_OTHER:
        p1->setExitFromPlayersHome(*p2);
        p2->setExitFromPlayersHome(*p1);
        break;
    default:
        dg->getExit(0, pos, facing);
        p1->setExit(pos, facing);
        dg->getExit(1, pos, facing);
        p2->setExit(pos, facing);
        break;
    }

    Mediator::instance().addPlayer(*p1);
    Mediator::instance().addPlayer(*p2);

    // add cheat controls (dview 1 only, for now)
    dviews[1]->addCheatControls(&sc.getCheatControls());
    
    // set up stuff manager.
    pimpl->stuff_manager.setStuffBagGraphic(sc.getStuffBagGraphic());
    
    // configure monsters and gore
    sc.configureGoreManager(pimpl->gore_manager, Mediator::instance().getPlayers());
    sc.configureMonsterManager(pimpl->monster_manager, dg->getVampireBats(),
                               dg->getZombieActivity());
    sc.configureEventManager(pimpl->event_manager);

    // add initial vampire bats:
    const int vbat_level = dg->getVampireBats();
    const int nbats_normal = vbat_level==0 ? 0 : 2*vbat_level + 1;
    const int nbats_guarded = vbat_level + 5;
    dg->addVampireBats(*pimpl->dmap, pimpl->monster_manager, nbats_normal, nbats_guarded);
    
    // we also have to set up a task that runs the monster manager every so often ...
    shared_ptr<MonsterTask> mtsk (new MonsterTask);
    pimpl->task_manager.addTask(mtsk, TP_LOW, pimpl->task_manager.getGVT()+1);
    
    // set "premapped" flag
    pimpl->premapped = dg->isPremapped();

    // set starting gears
    pimpl->gears = dg->getStartingGears();

    // Set up a task to respawn items.
    shared_ptr<ItemRespawnTask> itsk(new ItemRespawnTask(dg->getRespawnGenerator(),
                                                         dg->getRespawnLockpicks(),
                                                         dg->getRespawnInterval(),
                                                         dg->getRespawnTimeBase(),
                                                         dg->getRespawnOdds()));
    pimpl->task_manager.addTask(itsk, TP_NORMAL, pimpl->task_manager.getGVT()+1);
}

GameServer::~GameServer()
{
    // Clear the map out before deleting players. This makes sure that (for example)
    // Knight dtor does not attempt to call methods on the deleted Player objects.
    if (pimpl->dmap) {
        pimpl->dmap->clearAll();
        pimpl->dmap.reset();
    }
    Mediator::instance().deletePlayers();
    Mediator::destroyInstance();
}

void GameServer::spawnPlayers()
{
    const vector<Player*> &players(Mediator::instance().getPlayers());
    for (int i=0; i<players.size(); ++i) {
        // Transfer the starting gear from the dungeon-generator to the Player.
        for (int j=0; j<pimpl->gears.size(); ++j) {
            players[i]->addStartingGear(*pimpl->gears[j].first, pimpl->gears[j].second);
        }

        // (Re)spawn the player
        players[i]->respawn();

        // We insert premapping here, as well.
        if (pimpl->premapped) {
            ASSERT(players[i]->getKnight());
            MagicMapping(players[i]->getKnight());
        }
    }
}


//
// update
//

void GameServer::update(int new_gvt)
{
    // Do nothing if the game is won :)
    if (Mediator::instance().getWinningPlayer()) return;
    
    // update task manager
    pimpl->task_manager.advanceToTime(new_gvt);

    // 'update' all dungeonviews
    const vector<Player*> &players(Mediator::instance().getPlayers());
    for (vector<Player*>::const_iterator it = players.begin(); it != players.end(); ++it) {
        (*it)->getDungeonView().update();
    }
}

