/*
 * menu.hpp
 *
 * Dungeon Environment Customization Menu.
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#ifndef MENU_HPP
#define MENU_HPP

#include "boost/shared_ptr.hpp"
using namespace boost;

#include <string>
#include <vector>
using namespace std;

class DungeonDirective;
class Graphic;
class MenuConstraint;
class MenuItem;
class MenuSelections;
class MenuValue;
class Quest;


//
// Menu -- describes the layout of the menu itself. Contains MenuItems plus some
// DungeonDirectives that are always active.
//

class Menu {
public:
	// NB contained items, directives are NOT deleted upon destruction

	void addItem(const MenuItem *m)
		{ if (m) items.push_back(m); }
	void addDungeonDirective(const DungeonDirective *d)
		{ if (d) dungeon_directives.push_back(d); }

	int getNumItems() const { return items.size(); }
	const MenuItem & getItem(int n) const { return *(items.at(n)); }
	
	int getNumDungeonDirectives() const
		{ return dungeon_directives.size(); }
	const DungeonDirective * getDungeonDirective(int n) const
		{ return dungeon_directives.at(n); }

private:
	vector<const MenuItem *> items;
	vector<const DungeonDirective *> dungeon_directives;
};


//
// MenuItem -- a particular item in the menu. Contains a list of MenuValues, of which
// one is selected at any given time.
//

class MenuItem {
public:
	MenuItem() : min_value(0) { } 

	void setID(const string &i) { id = i; }
	void setTitle(const string &h) { title = h; }
	void setMinValue(int n) { min_value = n; }
	void addValue(const MenuValue *m) { if (m) values.push_back(m); }
	
	const string &getID() const { return id; }
	const string &getTitle() const { return title; }
	int getMinValue() const { return min_value; }
	int getNumValues() const { return values.size(); }
	const MenuValue & getValue(int n) const { return *(values.at(n - min_value)); }
	
private:
	int min_value;
	string id;
	string title;
	vector<const MenuValue *> values;
};


//
// MenuValue -- a possible selection for a MenuItem. Contains DungeonDirectives & other
// config info to be used when this value is selected.
//

class MenuValue {
public:
	MenuValue() : c_icon(0), g_icon(0) { }
	
	void setText(const string &s) { text = s; }
	void setCharIcon(char c) { c_icon = c; }
	void setGraphicIcon(const Graphic *g) { g_icon = g; }
	void addConstraint(shared_ptr<MenuConstraint> c) { if (c) constraints.push_back(c); }
	void addDungeonDirective(const DungeonDirective *d)
		{ if (d) dungeon_directives.push_back(d); }
	void addQuest(shared_ptr<Quest> q) { if (q) quests.push_back(q); }
	
	const string & getText() const { return text; }
	char getCharIcon() const { return c_icon; }
	const Graphic * getGraphicIcon() const { return g_icon; }

	int getNumDungeonDirectives() const { return dungeon_directives.size(); }
	const DungeonDirective *getDungeonDirective(int n) const
		{ return dungeon_directives.at(n); }
	int getNumQuests() const { return quests.size(); }
	shared_ptr<Quest> getQuest(int n) const { return quests.at(n); }

	// apply this value's constraints
	// returns true if something was changed.
	bool applyConstraints(MenuSelections &msel) const;
	
private:
	string text;
	char c_icon;
	const Graphic *g_icon;
	vector<const DungeonDirective *> dungeon_directives;
	vector<shared_ptr<Quest> > quests;
	vector<shared_ptr<MenuConstraint> > constraints;
};

#endif
