/*
 * gapi_display.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "app/gapi_app2d.hpp"
#include "display.hpp"
#include "gapi_graphic.hpp"
#include "misc/my_exceptions.hpp"

#include "boost/scoped_array.hpp"
using namespace boost;


//
// DisplayImpl struct for GAPI
//

struct DisplayImpl {
	int dpy_width, dpy_height;
	int xpitch, ypitch;
	int clip_x1, clip_x2, clip_y1, clip_y2;
	scoped_array<unsigned short> back_buffer;

	void doDrawGraphic(int bl_x, int bl_y, const Graphic &gfx, const ColourChange *cc);

	Graphic *font_graphic;
	enum { CHARPOS_MAX = 512 };
	int charpos[CHARPOS_MAX];
	void loadFont();
	void putString(int, int, const char *);
	int textWidth(const char *);
};


//
// Font handling stuff
// Much of this is based on SFont.c.
//

void DisplayImpl::loadFont()
{
	if (font_graphic) return;

	string font_name("knights_sfont.bmp");
	font_graphic = CreateGraphic(font_name, 0,0,0,0,0);
	if (!font_graphic) throw UnexpectedError(string("Could not load font (")+font_name+")");

	int x=0, i=0;
	if (!GapiApp2D::self()) throw UnexpectedError("app not yet open");
	const unsigned short int pink = GapiApp2D::self()->rgbToScreen(255,0,255);
	while (x < GetWidth(*font_graphic)) {
		if (font_graphic->getPixel(x,0) == pink) {
			if (i >= CHARPOS_MAX) break;
			charpos[i++] = x;
			while (x < GetWidth(*font_graphic)-1 && font_graphic->getPixel(x,0) == pink) {
				++x;
			}
			if (i >= CHARPOS_MAX) break;
			charpos[i++] = x;
		}
		++x;
	}
}

void DisplayImpl::putString(int bl_x, int bl_y, const char *text)
{
	int ofs;
	for (const char *p = text; *p != 0; ++p) {
		if (*p == ' ') {
			bl_x += charpos[2] - charpos[1];
		} else {
			ofs = (static_cast<unsigned char>(*p) - 33)*2 + 1;
			const int w = (charpos[ofs+2] + charpos[ofs+1]) / 2
				- (charpos[ofs]+charpos[ofs-1]) / 2;
			const int h = GetHeight(*font_graphic) - 1;
			const int sx = (charpos[ofs] + charpos[ofs-1]) / 2;
			const int dx = bl_x;
			const int dy = dpy_height-1-(bl_y+h-1);
			if (dx >= 0 && dx+w <= dpy_width && dy >= 0 && dy+h <= dpy_height) {
				font_graphic->blit(sx, 1, sx+w, h, dx, dy,
								   back_buffer.get(), dpy_width, 0);
			}
			bl_x += charpos[ofs+1] - charpos[ofs];
		}
	}
}

int DisplayImpl::textWidth(const char *txt)
{
	int ofs=0;
	int x=0;
	for (const char *p = txt; *p != 0; ++p) {
		if (*p == ' ') {
			x += charpos[2] - charpos[1];
		} else {
			ofs = (static_cast<unsigned char>(*p) - 33)*2+1;
			x += charpos[ofs+1] - charpos[ofs];
		}
	}
	return x;
}


//
// Display class for GAPI
//

Display::Display()
	: pimpl(new DisplayImpl)
{
	GapiApp2D * app = GapiApp2D::self();
	ASSERT(app);
	ASSERT(app->displayOpen());
	pimpl->dpy_width = app->getDisplayWidth();
	pimpl->dpy_height = app->getDisplayHeight();
	pimpl->xpitch = app->getXPitch();
	pimpl->ypitch = app->getYPitch();
	pimpl->clip_x1 = 0;
	pimpl->clip_y1 = 0;
	pimpl->clip_x2 = pimpl->dpy_width;
	pimpl->clip_y2 = pimpl->dpy_height;
	pimpl->back_buffer.reset(new unsigned short [pimpl->dpy_width * pimpl->dpy_height]);

	// Load the font
	pimpl->loadFont();
}

void Display::begin()
{
	// We have to clear the screen
	// 16-bpp is assumed
	memset(pimpl->back_buffer.get(), 0, pimpl->dpy_width * pimpl->dpy_height * 2);
}

void Display::end()
{
	// Do nothing if the app is not active!!
	if (!GapiApp2D::self()->isActive()) return;
	
	// Copy the back buffer to the actual video memory.
	unsigned short * vidmem = static_cast<unsigned short *>
		(GapiApp2D::self()->gapiBeginDraw());
	if (!vidmem) return;

	unsigned short *out_line = vidmem;
	const unsigned short *in_line = pimpl->back_buffer.get();
	for (int j=0; j<pimpl->dpy_height; ++j) {
		if (pimpl->xpitch == 2) {
			memcpy(out_line, in_line, pimpl->dpy_width*2);
		} else {
			const unsigned short *in = in_line;
			const unsigned short *in_stop = in_line + pimpl->dpy_width;
			unsigned short *out = out_line;
			while (in != in_stop) {
				*out = *in++;
				out += (pimpl->xpitch/2);
			}
		}
		out_line += (pimpl->ypitch/2);
		in_line += (pimpl->dpy_width);
	}
			
	GapiApp2D::self()->gapiEndDraw();
}

//
// get width/height
//

int Display::getWidth() const
{
	return pimpl->dpy_width;
}

int Display::getHeight() const
{
	return pimpl->dpy_height;
}

//
// clipping
//

void Display::setClipRectangle(int bl_x, int bl_y, int tr_x, int tr_y)
{
	pimpl->clip_x1 = bl_x;
	pimpl->clip_x2 = tr_x;
	pimpl->clip_y1 = pimpl->dpy_height - 1 - tr_y;
	pimpl->clip_y2 = pimpl->dpy_height - 1 - bl_y;
	if (pimpl->clip_x1 < 0) pimpl->clip_x1 = 0;
	if (pimpl->clip_y1 < 0) pimpl->clip_y1 = 0;
	if (pimpl->clip_x2 > pimpl->dpy_width) pimpl->clip_x2 = pimpl->dpy_width;
	if (pimpl->clip_y2 > pimpl->dpy_height) pimpl->clip_y2 = pimpl->dpy_height;
}

void Display::clearClipRectangle()
{
	pimpl->clip_x1 = 0;
	pimpl->clip_y1 = 0;
	pimpl->clip_x2 = pimpl->dpy_width;
	pimpl->clip_y2 = pimpl->dpy_height;
}

//
// Font/Text Display
//

void Display::drawText(int bl_x, int bl_y, const char *str)
{
	pimpl->putString(bl_x, bl_y, str);
}

int Display::getTextWidth(const char *str)
{
	return pimpl->textWidth(str);
}

//
// Graphic drawing
//

void Display::drawGraphic(int bl_x, int bl_y, const Graphic &gfx, bool)
{
	pimpl->doDrawGraphic(bl_x, bl_y, gfx, 0);
}

void Display::drawGraphicWithCC(int bl_x, int bl_y, const Graphic &gfx,
								const ColourChange &cc)
{
	pimpl->doDrawGraphic(bl_x, bl_y, gfx, &cc);
}

void DisplayImpl::doDrawGraphic(int bl_x, int bl_y, const Graphic &gfx, const ColourChange *cc)
{
	int sx1 = 0;
	int sy1 = 0;
	int sx2 = gfx.getWidth();
	int sy2 = gfx.getHeight();
	
	int dx = bl_x - gfx.getHX();
	int dy = this->dpy_height - bl_y - gfx.getHeight() + gfx.getHY();

	// Clipping
	if (dx < this->clip_x1) {
		sx1 += (this->clip_x1 - dx);
		dx = this->clip_x1;
	}
	if (dx + (sx2-sx1) >= this->clip_x2) {
		sx2 -= (dx + (sx2-sx1) - this->clip_x2 + 1);
	}
	if (sx1 >= sx2) return; // clipped out
	
	if (dy < this->clip_y1) {
		sy1 += (this->clip_y1 - dy);
		dy = this->clip_y1;
	}
	if (dy + (sy2-sy1) >= this->clip_y2) {
		sy2 -= (dy + (sy2-sy1) - this->clip_y2 + 1);
	}
	if (sy1 >= sy2) return; // clipped out

	gfx.blit(sx1, sy1, sx2, sy2, dx, dy, this->back_buffer.get(), this->dpy_width, cc);
}

//
// Pixel drawing
//

void Display::beginPixels()
{
}

void Display::drawPixel(int x, int y, int R, int G, int B)
{
	// Plot in the off-screen buffer
	// Note: 16 bits per pixel is assumed
	if (x < 0 || y < 0 || x >= pimpl->dpy_width || y >= pimpl->dpy_height) return;
	pimpl->back_buffer[x + (pimpl->dpy_height-1-y)*pimpl->dpy_width]
		= GapiApp2D::self()->rgbToScreen(R,G,B);
}

void Display::readPixel(int x, int y, int &R, int &G, int &B)
{
	if (x < 0 || y < 0 || x >= pimpl->dpy_width || y >= pimpl->dpy_height) return;
	GapiApp2D::self()->screenToRgb(
		pimpl->back_buffer[x + (pimpl->dpy_height-1-y)*pimpl->dpy_width],
		R, G, B);
}

void Display::endPixels()
{
}
