/*
 * sdl_graphic_cache.hpp
 *
 * This is used to cache colour-changed and resized graphics.
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#ifndef SDL_GRAPHIC_CACHE_HPP
#define SDL_GRAPHIC_CACHE_HPP

#include "colour_change.hpp"

#include <list>
#include <map>
using namespace std;

class Display;
class Graphic;

class SdlGraphicCache {
public:
    SdlGraphicCache();
    ~SdlGraphicCache();

    // set the Display pointer
    void setDisplay(Display &d) { display = &d; }
    
    // Input: an 'original' graphic
    // Output: a new graphic with the colours changed
    // Note: All graphics should be used immediately before getting any other graphic from
    // the cache, since the getGraphic routines can delete cached graphics.
    // All cached graphics are deleted in the dtor.
    const Graphic * getGraphicWithCC(const Graphic &original, const ColourChange &cc);

    // Create a new resized graphic
    const Graphic * getResizedGraphic(const Graphic &original, int new_width, int new_height);

    // Colour-change and resize together
    const Graphic * getResizedGraphicWithCC(const Graphic &original, int new_width, int new_height, const ColourChange &cc);

    // Wipe out the caches. (Used on resizer change)
    void clear();
    
private:
    SdlGraphicCache(const SdlGraphicCache&);  // not defined
    void operator=(const SdlGraphicCache&);   // not defined

    struct GraphicKey;
    struct GraphicData;
    
    typedef map<GraphicKey, GraphicData> GraphicMap;
    typedef list<GraphicKey> RecentList;  // MRU at front, LRU at back
    
    struct GraphicKey {
        const Graphic * original;
        ColourChange cc;
        int new_width, new_height;

        bool operator<(const GraphicKey &rhs) const
        {
            return
                original < rhs.original ? true :
                rhs.original < original ? false :
                cc < rhs.cc ? true :
                rhs.cc < cc ? false :
                new_width < rhs.new_width ? true :
                rhs.new_width < new_width ? false :
                new_height < rhs.new_height;
        }
    };

    struct GraphicData {
        const Graphic * new_graphic;
        RecentList::iterator list_position;
    };

    
    // Data
    
    GraphicMap graphic_map;
    RecentList recent_list;
    
    int npix;
        
    Display *display;

    
    // Functions

    const Graphic * createGraphic(const GraphicKey &key);
    const Graphic * getGraphic(const GraphicKey &key);
    bool tooBig() const;
    void deleteOld();    
};

// Global access point
extern SdlGraphicCache g_sdl_graphic_cache;

#endif
